from unittest import TestCase

from pyramid.testing import DummyRequest
from pyramid.exceptions import ConfigurationError

__all__ = [
    'ConfigurationError', 'Dummy', 'DummyConfig', 'DummyLdap3',
    'DummyLDAPConnector', 'DummyManager', 'DummyRequest', 'DummySearch',
    'TestCase']


class DummyLDAPConnector(object):

    def __init__(self, group_list):
        self.group_list = group_list

    # noinspection PyUnusedLocal
    def user_groups(self, userdn):
        return self.group_list


class Dummy(object):

    def __init__(self, *arg, **kw):
        pass


class DummyConfig(object):

    introspectable = Dummy

    def __init__(self):
        self.registry = Dummy()
        self.directives = []
        self.req_method = self.req_method_args = None

    # noinspection PyUnusedLocal
    def add_directive(self, name, directive):
        self.directives.append(name)

    def add_request_method(self, method, name, property=False, reify=False):
        self.req_method = method
        self.req_method_args = name, property, reify

    # noinspection PyUnusedLocal
    @staticmethod
    def action(discriminator, action_task, introspectables=()):
        if action_task:
            action_task()


class DummyManager(object):

    def __init__(self, with_error=None, with_result=None):
        self.with_error = with_error
        self.with_result = with_result
        self.user = self.password = None
        self.bound = None
        self.result_id = 0
        self.search_args = self.search_kwargs = None

    def __enter__(self):
        return self

    def __exit__(self, *_args):
        self.unbind()

    def connection(self, user=None, password=None):
        self.user = user
        self.password = password
        if self.with_error:
            raise self.with_error
        self.bind()
        return self

    def bind(self):
        self.bound = True
        return False

    def unbind(self):
        self.bound = False

    def search(self, *args, **kwargs):
        if not self.bound:  # pragma: no cover
            raise AssertionError('connection not yet bound')
        self.search_args = args
        self.search_kwargs = kwargs
        self.result_id += 1
        return self.result_id

    def get_response(self, result_id):
        if self.search_kwargs['search_scope'] == 'none':
            return None, result_id
        return self.with_result, result_id


class DummySearch(object):

    def __init__(self, result, exc=None):
        self.result = result
        self.exc = exc
        self.manager = None
        self.kw = None

    def execute(self, manager, **kw):
        if self.exc is not None:
            raise self.exc
        self.manager = manager
        self.kw = kw
        return self.result


class DummyLdap3Server(object):

    def __init__(
            self, host, port=None, use_ssl=False, tls=None, get_info=None):
        self.host = host
        self.port = port
        self.ssl = use_ssl
        self.tls = tls
        self.get_info = get_info


class DummyLdap3Connection(object):

    def __init__(
            self, server, user=None, password=None,
            auto_bind=False, authentication=None,
            read_only=False, client_strategy=None, lazy=False,
            pool_name='pyramid_ldap3', pool_size=10, pool_lifetime=None):
        self.server = server
        self.user = user
        self.password = password
        self.auto_bind = auto_bind
        self.authentication = authentication
        self.read_only = read_only
        self.client_strategy = client_strategy
        self.lazy = lazy
        self.pool_name = pool_name
        self.pool_size = pool_size
        self.pool_lifetime = pool_lifetime


class DummyLdap3(object):

    ASYNC = 1
    REUSABLE = 2

    NONE = 'NO_INFO'
    SCHEMA = 'SCHEMA'
    ALL = 'ALL'

    Server = DummyLdap3Server
    Connection = DummyLdap3Connection



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# http://www.sphinx-doc.org/en/master/config

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
# import os
# import sys
# sys.path.insert(0, os.path.abspath('.'))

import pylons_sphinx_themes

# -- Project information -----------------------------------------------------

project = 'pyramid_ldap3'
copyright = '2012-2020, Chris McDonough, Christoph Zwerschke'
author = 'Chris McDonough, Christoph Zwerschke'

# The short X.Y version
version = '0.4'
# The full version, including alpha/beta/rc tags
release = '0.4.2'


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = ['sphinx.ext.autodoc']

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']


# -- Options for HTML output -------------------------------------------------

# Add and use Pyramid theme
html_theme = 'pyramid'
html_theme_path = pylons_sphinx_themes.get_html_themes_path()
html_theme_options = dict(
    canonical_url='https://pyramid-ldap3.readthedocs.io/',
    github_url='https://github.com/Cito/pyramid_ldap3')

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']



import ldap3

from pyramid_ldap3 import groupfinder

from pyramid.config import Configurator
from pyramid.authentication import AuthTktAuthenticationPolicy
from pyramid.authorization import ACLAuthorizationPolicy
from pyramid.security import Allow, Authenticated


class RootFactory(object):

    __acl__ = [(Allow, Authenticated, 'view')]

    def __init__(self, request):
        pass


def main(global_config, **settings):
    config = Configurator(settings=settings, root_factory=RootFactory)
    config.include('pyramid_ldap3')
    config.include('pyramid_chameleon')
    config.set_authentication_policy(
        AuthTktAuthenticationPolicy('seekr1t', callback=groupfinder))
    config.set_authorization_policy(
        ACLAuthorizationPolicy())
    config.ldap_setup(
        'ldap://ldap.forumsys.com',
        bind='cn=read-only-admin,dc=example,dc=com',
        passwd='password')
    config.ldap_set_login_query(
        'dc=example,dc=com',
        # '(sAMAccountName=%(login)s)',
        '(uid=%(login)s)',
        scope=ldap3.SUBTREE,
        cache_period=0)
    config.ldap_set_groups_query(
        'dc=example,dc=com',
        # '(member:1.2.840.113556.1.4.1941:=%(userdn)s)',
        # '(&(objectCategory=group)(member=%(userdn)s))',
        '(&(objectClass=groupOfUniqueNames)(uniqueMember=%(userdn)s))',
        cache_period=60)
    config.add_route('sampleapp.root', '/')
    config.add_route('sampleapp.login', '/login')
    config.add_route('sampleapp.logout', '/logout')
    config.scan('.views')
    return config.make_wsgi_app()



from . import TestCase, DummyConfig, DummyRequest


class TestLdapSetup(TestCase):

    def _call_fut(self, config, uri, **kw):
        from pyramid_ldap3 import ldap_setup
        return ldap_setup(config, uri, **kw)

    def test_it_defaults_ldap_host(self):
        from pyramid_ldap3 import Connector
        config = DummyConfig()
        self._call_fut(config, 'ldap://dummyhost')
        self.assertEqual(config.req_method_args,
                         ('ldap_connector', True, True))
        request = DummyRequest()
        connector = config.req_method(request)
        self.assertEqual(connector.__class__, Connector)
        server = connector.manager.server
        import ldap3
        self.assertIsInstance(server, ldap3.Server)
        self.assertEqual(server.host, 'dummyhost')
        self.assertEqual(server.port, 389)
        self.assertFalse(server.tls)
        self.assertEqual(server.get_info, ldap3.NONE)

    def test_it_defaults_ldaps_host(self):
        from pyramid_ldap3 import Connector
        config = DummyConfig()
        self._call_fut(config, 'ldaps://dummyhost')
        self.assertEqual(config.req_method_args,
                         ('ldap_connector', True, True))
        request = DummyRequest()
        connector = config.req_method(request)
        self.assertEqual(connector.__class__, Connector)
        server = connector.manager.server
        import ldap3
        self.assertIsInstance(server, ldap3.Server)
        self.assertEqual(server.host, 'dummyhost')
        self.assertEqual(server.port, 636)
        self.assertTrue(server.tls)
        self.assertEqual(server.get_info, ldap3.NONE)

    def test_it_defaults_ldap_hosts(self):
        from pyramid_ldap3 import Connector
        config = DummyConfig()
        self._call_fut(config, (
            'ldap://plainhost', 'ldaps://sslhost', 'ldap://custom:8389'))
        self.assertEqual(config.req_method_args,
                         ('ldap_connector', True, True))
        request = DummyRequest()
        connector = config.req_method(request)
        self.assertEqual(connector.__class__, Connector)
        server_pool = connector.manager.server
        import ldap3
        self.assertIsInstance(server_pool, ldap3.ServerPool)
        self.assertEqual(len(server_pool), 3)
        server = server_pool[0]
        self.assertEqual(server.host, 'plainhost')
        self.assertEqual(server.port, 389)
        self.assertFalse(server.tls)
        self.assertEqual(server.get_info, ldap3.NONE)
        server = server_pool[1]
        self.assertEqual(server.host, 'sslhost')
        self.assertEqual(server.port, 636)
        self.assertTrue(server.tls)
        self.assertEqual(server.get_info, ldap3.NONE)
        server = server_pool[2]
        self.assertEqual(server.host, 'custom')
        self.assertEqual(server.port, 8389)
        self.assertFalse(server.tls)
        self.assertEqual(server.get_info, ldap3.NONE)

    def test_it_defaults_realm(self):
        from pyramid_ldap3 import Connector
        config = DummyConfig()
        self._call_fut(config, 'ldap://dummyhost', realm='test')
        self.assertEqual(config.req_method_args,
                         ('ldap_connector_test', True, True))
        request = DummyRequest()
        connector = config.req_method(request)
        self.assertEqual(connector.__class__, Connector)
        server = connector.manager.server
        import ldap3
        self.assertIsInstance(server, ldap3.Server)
        self.assertEqual(server.host, 'dummyhost')
        self.assertEqual(server.port, 389)
        self.assertFalse(server.tls)
        self.assertEqual(server.get_info, ldap3.NONE)




from pyramid.view import view_config, forbidden_view_config
from pyramid.response import Response
from pyramid.httpexceptions import HTTPFound
from pyramid.security import remember, forget

from pyramid_ldap3 import get_ldap_connector


@view_config(route_name='sampleapp.root', permission='view')
def logged_in(request):
    return Response(
        'OK (%s)' % ', '.join(request.effective_principals))


@view_config(route_name='sampleapp.logout')
def logout(request):
    headers = forget(request)
    return Response('Logged out', headers=headers)


@view_config(route_name='sampleapp.login', renderer='templates/login.pt')
@forbidden_view_config(renderer='templates/login.pt')
def login(request):
    url = request.current_route_url()
    login = ''
    password = ''
    error = ''

    if 'form.submitted' in request.POST:
        login = request.POST['login']
        password = request.POST['password']
        connector = get_ldap_connector(request)
        data = connector.authenticate(login, password)
        if data is not None:
            dn = data[0]
            headers = remember(request, dn)
            return HTTPFound('/', headers=headers)
        error = 'Invalid credentials'

    return dict(
        login_url=url,
        login=login,
        password=password,
        error=error)



from . import TestCase, DummyLdap3


class TestConnectionManager(TestCase):

    def _make_one(
            self, uri,
            bind=None, passwd=None, tls=None,
            use_pool=True, pool_size=10, pool_lifetime=3600,
            get_info=None, realm=None):
        from pyramid_ldap3 import ConnectionManager
        ldap3 = DummyLdap3()
        return ConnectionManager(
            uri, bind=bind, passwd=passwd, tls=tls,
            use_pool=use_pool, pool_size=pool_size,
            pool_lifetime=pool_lifetime,
            get_info=get_info, ldap3=ldap3, realm=realm)

    def test_uri(self):
        manager = self._make_one('testhost')
        self.assertFalse(manager.server.ssl)
        self.assertEqual(manager.server.host, 'testhost')
        self.assertEqual(manager.server.port, 389)
        manager = self._make_one('ldap://testhost')
        self.assertFalse(manager.server.ssl)
        self.assertEqual(manager.server.host, 'testhost')
        self.assertEqual(manager.server.port, 389)
        manager = self._make_one('ldap://testhost:432')
        self.assertFalse(manager.server.ssl)
        self.assertEqual(manager.server.host, 'testhost')
        self.assertEqual(manager.server.port, 432)
        manager = self._make_one('ldaps://testhost')
        self.assertTrue(manager.server.ssl)
        self.assertEqual(manager.server.host, 'testhost')
        self.assertEqual(manager.server.port, 636)
        manager = self._make_one('ldaps://testhost:987')
        self.assertTrue(manager.server.ssl)
        self.assertEqual(manager.server.host, 'testhost')
        self.assertEqual(manager.server.port, 987)

    def test_use_tls(self):
        manager = self._make_one('testhost')
        self.assertFalse(manager.server.tls)
        from pyramid_ldap3 import ldap3
        tls = ldap3.Tls()
        manager = self._make_one('testhost', tls=tls)
        self.assertIs(manager.server.tls, tls)

    def test_get_info(self):
        manager = self._make_one('testhost')
        from pyramid_ldap3 import ldap3
        self.assertEqual(manager.server.get_info, ldap3.NONE)
        manager = self._make_one('testhost', get_info=ldap3.ALL)
        self.assertEqual(manager.server.get_info, ldap3.ALL)

    def test_bind(self):
        manager = self._make_one('testhost', 'fred', 'flint')
        self.assertEqual(manager.bind, 'fred')
        self.assertEqual(manager.passwd, 'flint')

    def test_pool(self):
        manager = self._make_one('testhost')
        self.assertEqual(manager.pool_size, 10)
        manager = self._make_one('testhost', pool_size=42)
        self.assertEqual(manager.pool_size, 42)
        manager = self._make_one('testhost', pool_lifetime=4200)
        self.assertEqual(manager.pool_lifetime, 4200)
        manager = self._make_one('testhost', use_pool=False)
        self.assertIsNone(manager.pool_size)

    def test_connection(self):
        manager = self._make_one('testhost')
        conn = manager.connection()
        self.assertEqual(conn.server.host, 'testhost')
        self.assertIsNone(conn.user)
        conn = manager.connection('fred', 'flint')
        self.assertEqual(conn.user, 'fred')
        self.assertEqual(conn.password, 'flint')

    def test_connection_with_realm(self):
        manager = self._make_one('testhost', realm='test_realm')
        conn = manager.connection()
        self.assertEqual(conn.server.host, 'testhost')
        self.assertIsNone(conn.user)
        conn = manager.connection('fred', 'flint')
        self.assertEqual(conn.user, 'fred')
        self.assertEqual(conn.password, 'flint')



from . import (
    TestCase, ConfigurationError,
    Dummy, DummyRequest, DummyManager, DummySearch)


class TestGetLdapConnector(TestCase):

    def _call_fut(self, request, realm=None):
        from pyramid_ldap3 import get_ldap_connector
        return get_ldap_connector(request, realm)

    def test_no_connector(self):
        request = DummyRequest()
        self.assertRaises(ConfigurationError, self._call_fut, request)

    def test_with_connector(self):
        request = DummyRequest()
        request.ldap_connector = True
        result = self._call_fut(request)
        self.assertEqual(result, True)

    def test_with_connector_and_realm(self):
        request = DummyRequest()
        request.ldap_connector_test_realm = True
        self.assertRaises(ConfigurationError, self._call_fut, request)
        result = self._call_fut(request, 'test_realm')
        self.assertEqual(result, True)


class TestConnectorWithoutRealm(TestCase):

    def _make_one(self, registry, manager):
        from pyramid_ldap3 import Connector
        return Connector(registry, manager)

    def test_authenticate_no_ldap_login_query(self):
        manager = DummyManager()
        inst = self._make_one(None, manager)
        self.assertRaises(ConfigurationError, inst.authenticate, None, None)
        self.assertIsNone(manager.bound)

    def test_authenticate_search_returns_no_result(self):
        manager = DummyManager()
        registry = Dummy()
        registry.ldap_login_query = DummySearch([])
        inst = self._make_one(registry, manager)
        self.assertIsNone(inst.authenticate(None, None))
        self.assertIsNone(manager.bound)

    def test_authenticate_empty_password(self):
        manager = DummyManager()
        registry = Dummy()
        registry.ldap_login_query = DummySearch([('a', 'b')])
        inst = self._make_one(registry, manager)
        self.assertIsNone(inst.authenticate('foo', ''))
        self.assertIsNone(manager.bound)

    def test_authenticate_search_returns_one_result(self):
        manager = DummyManager()
        registry = Dummy()
        registry.ldap_login_query = DummySearch([('a', 'b')])
        inst = self._make_one(registry, manager)
        self.assertEqual(inst.authenticate(None, None), ('a', 'b'))
        self.assertIs(manager.bound, False)

    def test_authenticate_search_returns_multiple_results(self):
        manager = DummyManager()
        registry = Dummy()
        registry.ldap_login_query = DummySearch([('a', 'b'), ('a', 'c')])
        inst = self._make_one(registry, manager)
        self.assertIsNone(inst.authenticate(None, None))
        self.assertIsNone(manager.bound)

    def test_authenticate_search_bind_raises(self):
        from pyramid_ldap3 import LDAPException
        manager = DummyManager(with_error=LDAPException)
        registry = Dummy()
        registry.ldap_login_query = DummySearch([('a', 'b')])
        inst = self._make_one(registry, manager)
        self.assertIsNone(inst.authenticate(None, None))
        self.assertIsNone(manager.bound)

    def test_authenticate_search_escapes(self):
        manager = DummyManager()
        registry = Dummy()
        search = DummySearch([('a', 'b')])
        registry.ldap_login_query = search
        inst = self._make_one(registry, manager)
        self.assertEqual(inst.authenticate('abc123', 'def456'), ('a', 'b'))
        self.assertEqual(search.kw['login'], 'abc123')
        self.assertEqual(search.kw['password'], 'def456')
        self.assertEqual(inst.authenticate('abc*', 'def*'), ('a', 'b'))
        self.assertEqual(search.kw['login'], 'abc\\2A')
        self.assertEqual(search.kw['password'], 'def\\2A')
        self.assertEqual(inst.authenticate(b'ab\xc3\xa7123', None), ('a', 'b'))
        self.assertEqual(search.kw['login'].encode('latin-1'), b'ab\xe7123')
        self.assertEqual(inst.authenticate(b'ab\xe7123', None), ('a', 'b'))
        self.assertEqual(search.kw['login'], '\\61\\62\\e7\\31\\32\\33')
        self.assertIs(manager.bound, False)

    def test_user_groups_no_ldap_groups_query(self):
        manager = DummyManager()
        inst = self._make_one(None, manager)
        self.assertRaises(ConfigurationError, inst.user_groups, None)
        self.assertIsNone(manager.bound)

    def test_user_groups_search_returns_result(self):
        manager = DummyManager()
        registry = Dummy()
        registry.ldap_groups_query = DummySearch([('a', 'b')])
        inst = self._make_one(registry, manager)
        self.assertEqual(inst.user_groups(None), [('a', 'b')])
        self.assertIsNone(manager.bound)

    def test_user_groups_execute_raises(self):
        from pyramid_ldap3 import LDAPException
        manager = DummyManager()
        registry = Dummy()
        registry.ldap_groups_query = DummySearch(
            [('a', 'b')], LDAPException)
        inst = self._make_one(registry, manager)
        self.assertIsNone(inst.user_groups(None))
        self.assertIsNone(manager.bound)

    def test_user_groups_search_escapes(self):
        manager = DummyManager()
        registry = Dummy()
        search = DummySearch([('a', 'b')])
        registry.ldap_groups_query = search
        inst = self._make_one(registry, manager)
        self.assertEqual(inst.user_groups('abc123'), [('a', 'b')])
        self.assertEqual(search.kw['userdn'], 'abc123')
        self.assertEqual(inst.user_groups('(abc*123)'), [('a', 'b')])
        self.assertEqual(search.kw['userdn'], '\\28abc\\2A123\\29')
        self.assertEqual(inst.user_groups(b'ab\xc3\xa7123'), [('a', 'b')])
        self.assertEqual(search.kw['userdn'].encode('latin-1'), b'ab\xe7123')
        self.assertEqual(inst.user_groups(b'ab\xe7123'), [('a', 'b')])
        self.assertEqual(search.kw['userdn'], '\\61\\62\\e7\\31\\32\\33')
        self.assertIsNone(manager.bound)


class TestConnectorWithRealm(TestCase):

    def _make_one(self, registry, manager, realm=None):
        from pyramid_ldap3 import Connector
        return Connector(registry, manager, realm)

    def test_authenticate_no_ldap_login_query(self):
        manager = DummyManager()
        inst = self._make_one(None, manager, 'test_realm')
        self.assertRaises(ConfigurationError, inst.authenticate, None, None)
        self.assertIsNone(manager.bound)

    def test_authenticate_search_returns_correct_results(self):
        manager = DummyManager()
        registry = Dummy()
        registry.ldap_login_query_test_realm = DummySearch([('a', 'b')])
        registry.ldap_login_query_another_realm = DummySearch([('c', 'd')])
        inst = self._make_one(None, manager)
        self.assertRaises(ConfigurationError, inst.authenticate, None, None)
        self.assertIsNone(manager.bound)
        inst = self._make_one(registry, manager, 'test_realm')
        self.assertEqual(inst.authenticate(None, None), ('a', 'b'))
        self.assertIs(manager.bound, False)
        inst = self._make_one(registry, manager, 'another_realm')
        self.assertEqual(inst.authenticate(None, None), ('c', 'd'))
        self.assertIs(manager.bound, False)

    def test_user_groups_no_ldap_groups_query(self):
        manager = DummyManager()
        inst = self._make_one(None, manager, 'test_realm')
        self.assertRaises(ConfigurationError, inst.user_groups, None)
        self.assertIsNone(manager.bound)

    def test_user_groups_search_returns_correct_results(self):
        manager = DummyManager()
        registry = Dummy()
        registry.ldap_groups_query_test_realm = DummySearch([('a', 'b')])
        registry.ldap_groups_query_another_realm = DummySearch([('c', 'd')])
        inst = self._make_one(None, manager)
        self.assertRaises(ConfigurationError, inst.user_groups, None)
        self.assertIsNone(manager.bound)
        inst = self._make_one(registry, manager, 'test_realm')
        self.assertEqual(inst.user_groups(None), [('a', 'b')])
        self.assertIsNone(manager.bound)
        inst = self._make_one(registry, manager, 'another_realm')
        self.assertEqual(inst.user_groups(None), [('c', 'd')])
        self.assertIsNone(manager.bound)



from . import TestCase, DummyManager


class TestLDAPQuery(TestCase):

    def _make_one(self, base_dn, filter_tmpl, scope, attributes, cache_period):
        from pyramid_ldap3 import _LDAPQuery
        return _LDAPQuery(
            base_dn, filter_tmpl, scope, attributes, cache_period)

    def _manager(self, with_result):
        return DummyManager(with_result=with_result)

    def test_execute_no_result(self):
        inst = self._make_one('DN=Org', '(cn=%(login)s)', 'none', 'attrs', 0)
        manager = self._manager([{'dn': 'a', 'attributes': {'b': 'c'}}])
        result = inst.execute(manager, login='foo')
        self.assertEqual(result, [])

    def test_query_cache_no_rollover(self):
        inst = self._make_one(None, None, None, None, 1)
        inst.last_timeslice = 1 << 31
        inst.cache['foo'] = 'bar'
        self.assertEqual(inst.query_cache('foo'), 'bar')

    def test_query_cache_with_rollover(self):
        inst = self._make_one(None, None, None, None, 1)
        inst.cache['foo'] = 'bar'
        self.assertIsNone(inst.query_cache('foo'))
        self.assertEqual(inst.cache, {})
        self.assertNotEqual(inst.last_timeslice, 0)

    def test_execute_no_cache_period(self):
        inst = self._make_one('DN=Org', '(cn=%(login)s)', 'scope', 'attrs', 0)
        manager = self._manager([{'dn': 'a', 'attributes': {'b': 'c'}}])
        result = inst.execute(manager, login='foo')
        self.assertEqual(inst.cache, {})
        self.assertEqual(result, [('a', {'b': 'c'})])
        self.assertEqual(manager.search_args, ('DN=Org', '(cn=foo)'))
        self.assertEqual(manager.search_kwargs, {
            'attributes': 'attrs', 'search_scope': 'scope'})

    def test_execute_with_cache_period_miss(self):
        inst = self._make_one('DN=Org', '(cn=%(login)s)', 'scope', 'attrs', 1)
        manager = self._manager([{'dn': 'a', 'attributes': {'b': 'c'}}])
        result = inst.execute(manager, login='foo')
        self.assertEqual(result, [('a', {'b': 'c'})])
        self.assertEqual(manager.search_args, ('DN=Org', '(cn=foo)'))
        self.assertEqual(manager.search_kwargs, {
            'attributes': 'attrs', 'search_scope': 'scope'})

    def test_execute_with_cache_period_hit(self):
        inst = self._make_one('DN=Org', '(cn=%(login)s)', 'scope', 'attrs', 1)
        inst.last_timeslice = 1 << 31
        inst.cache[('DN=Org', '(cn=foo)')] = ('d', {'e': 'f'})
        manager = self._manager([{'dn': 'a', 'attributes': {'b': 'c'}}])
        result = inst.execute(manager, login='foo')
        self.assertEqual(result, ('d', {'e': 'f'}))
        self.assertIsNone(manager.search_args)
        self.assertIsNone(manager.search_kwargs)



from . import TestCase, DummyConfig


class TestIncludeme(TestCase):

    def _call_fut(self, config):
        from pyramid_ldap3 import includeme
        includeme(config)

    def test_it(self):
        config = DummyConfig()
        self._call_fut(config)
        self.assertEqual(config.directives, [
            'ldap_setup', 'ldap_set_login_query', 'ldap_set_groups_query'])



from . import TestCase, DummyConfig


class TestLdapSetGroupsQuery(TestCase):

    def _call_fut(self, config, base_dn, filter_tmpl, **kw):
        from pyramid_ldap3 import ldap_set_groups_query
        return ldap_set_groups_query(config, base_dn, filter_tmpl, **kw)

    def test_it_defaults(self):
        import ldap3
        config = DummyConfig()
        self._call_fut(config, 'dn', 'tmpl')
        ldap_groups_query = getattr(config.registry, 'ldap_groups_query', None)
        self.assertIsNotNone(ldap_groups_query)
        self.assertEqual(ldap_groups_query.base_dn, 'dn')
        self.assertEqual(ldap_groups_query.filter_tmpl, 'tmpl')
        self.assertEqual(ldap_groups_query.scope, ldap3.SUBTREE)
        self.assertEqual(ldap_groups_query.cache_period, 0)

    def test_it_realm(self):
        import ldap3
        config = DummyConfig()
        self._call_fut(config, 'dn_test', 'tmpl_test', realm='test')
        ldap_groups_query = getattr(
            config.registry, 'ldap_groups_query_test', None)
        self.assertIsNotNone(ldap_groups_query)
        self.assertEqual(ldap_groups_query.base_dn, 'dn_test')
        self.assertEqual(ldap_groups_query.filter_tmpl, 'tmpl_test')
        self.assertEqual(ldap_groups_query.scope, ldap3.SUBTREE)
        self.assertEqual(ldap_groups_query.cache_period, 0)


class TestLdapSetLoginQuery(TestCase):

    def _call_fut(self, config, base_dn, filter_tmpl, **kw):
        from pyramid_ldap3 import ldap_set_login_query
        return ldap_set_login_query(config, base_dn, filter_tmpl, **kw)

    def test_it_defaults(self):
        from pyramid_ldap3 import ldap3
        config = DummyConfig()
        self._call_fut(config, 'dn', 'tmpl')
        ldap_login_query = getattr(config.registry, 'ldap_login_query', None)
        self.assertIsNotNone(ldap_login_query)
        self.assertEqual(ldap_login_query.base_dn, 'dn')
        self.assertEqual(ldap_login_query.filter_tmpl, 'tmpl')
        self.assertEqual(ldap_login_query.scope, ldap3.LEVEL)
        self.assertEqual(ldap_login_query.cache_period, 0)

    def test_it_realm(self):
        from pyramid_ldap3 import ldap3
        config = DummyConfig()
        self._call_fut(config, 'dn_test', 'tmpl_test', realm='test')
        ldap_login_query = getattr(
            config.registry, 'ldap_login_query_test', None)
        self.assertIsNotNone(ldap_login_query)
        self.assertEqual(ldap_login_query.base_dn, 'dn_test')
        self.assertEqual(ldap_login_query.filter_tmpl, 'tmpl_test')
        self.assertEqual(ldap_login_query.scope, ldap3.LEVEL)
        self.assertEqual(ldap_login_query.cache_period, 0)




import logging

from time import time

from pyramid.exceptions import ConfigurationError

try:
    import ldap3
except ImportError:  # pragma: no cover
    # this is for benefit of being able to build the docs on rtd.org
    class _Ldap3Module(object):
        BASE = None
        LEVEL = None
        SUBTREE = None
        REUSABLE = None
    ldap3 = _Ldap3Module()
    LDAPException = Exception
else:
    LDAPException = ldap3.core.exceptions.LDAPException

logger = logging.getLogger(__name__)

_ord = ord if str is bytes else int

_escape_for_search = {
    '*': '\\2A', '(': '\\28', ')': '\\29', '\\': '\\5C', '\0': '\\00'}

__all__ = ['get_ldap_connector', 'get_groups', 'groupfinder']


def escape_for_search(s):
    """Escape search string for LDAP according to RFC4515 when necessary."""
    if not s:
        return s
    if isinstance(s, bytes):
        try:
            s = s.decode('utf-8')
        except UnicodeDecodeError:
            return ''.join('\\%02x' % _ord(b) for b in s)
    return ''.join((_escape_for_search.get(c, c) for c in s))


class _LDAPQuery(object):
    """Represents an LDAP query.

    Provides rudimentary in-RAM caching of query results.
    """

    def __init__(self, base_dn, filter_tmpl, scope, attributes, cache_period):
        self.base_dn = base_dn
        self.filter_tmpl = filter_tmpl
        self.scope = scope
        self.attributes = attributes
        self.cache_period = cache_period
        self.last_timeslice = 0
        self.cache = {}

    def __str__(self):
        return ('base_dn={base_dn}, filter_tmpl={filter_tmpl}, '
                'scope={scope}, attributes={attributes}, '
                'cache_period={cache_period}'.format(**self.__dict__))

    def query_cache(self, cache_key):
        now = time()
        ts = _timeslice(self.cache_period, now)

        if ts > self.last_timeslice:
            logger.debug(
                'dumping cache; now ts: %r, last_ts: %r',
                ts, self.last_timeslice)
            self.cache = {}
            self.last_timeslice = ts

        return self.cache.get(cache_key)

    def execute(self, manager, **kw):
        cache_key = (self.base_dn % kw, self.filter_tmpl % kw)

        logger.debug('searching for %r', cache_key)

        result = self.query_cache(cache_key) if self.cache_period else None
        if result is None:
            with manager.connection() as conn:
                ret = conn.search(
                    search_scope=self.scope,
                    attributes=self.attributes, *cache_key)
                result, ret = conn.get_response(ret)
            if result is None:
                result = []
            else:
                result = [(r['dn'], r['attributes']) for r in result
                          if 'dn' in r]
                if self.cache_period:
                    self.cache[cache_key] = result
        else:
            logger.debug('result for %r retrieved from cache', cache_key)

        logger.debug('search result: %r', result)

        return result


def _timeslice(period, when=None):
    if when is None:  # pragma: no cover
        when = time()
    return when - (when % period)


def _add_realm(base_identifier, realm=None):
    return '{}_{}'.format(base_identifier, realm) if realm else base_identifier


class ConnectionManager(object):
    """Provides API methods for managing LDAP connections."""

    # noinspection PyShadowingNames
    def __init__(
            self, uri, bind=None, passwd=None, tls=None,
            use_pool=True, pool_size=10, pool_lifetime=3600,
            get_info=None, ldap3=ldap3, realm=None):
        self.ldap3 = ldap3
        uris = uri if isinstance(uri, (list, tuple)) else uri.split()
        self.uri = uri[0] if len(uris) == 1 else uris
        if get_info is None:
            get_info = ldap3.NONE
        servers = []
        for uri in uris:
            try:
                schema, host = uri.split('://', 1)
            except ValueError:
                schema, host = 'ldap', uri
            use_ssl = schema == 'ldaps'
            try:
                host, port = host.split(':', 1)
                port = int(port)
            except ValueError:
                host, port = host, 636 if use_ssl else 389
            server = self.ldap3.Server(
                host, port=port, use_ssl=use_ssl, tls=tls,
                get_info=get_info)
            servers.append(server)
        self.server = servers[
            0] if len(servers) == 1 else self.ldap3.ServerPool(servers)
        self.bind, self.passwd = bind, passwd
        if use_pool:
            self.strategy = ldap3.REUSABLE
            self.pool_name = _add_realm('pyramid_ldap3', realm)
            self.pool_size = pool_size
            self.pool_lifetime = pool_lifetime
        else:
            self.strategy = ldap3.ASYNC
            self.pool_name = self.pool_size = self.pool_lifetime = None

    def __str__(self):
        return ('uri={uri}, bind={bind}/{passwd},pool={pool_size}'.format(
            **self.__dict__))

    def connection(self, user=None, password=None):
        if user:
            conn = self.ldap3.Connection(
                self.server, user=user, password=password,
                client_strategy=ldap3.SYNC,
                auto_bind=True, lazy=False, read_only=True)
        else:
            conn = self.ldap3.Connection(
                self.server, user=self.bind, password=self.passwd,
                client_strategy=self.strategy,
                pool_name=self.pool_name, pool_size=self.pool_size,
                pool_lifetime=self.pool_lifetime,
                auto_bind=True, lazy=False, read_only=True)
        return conn


class Connector(object):
    """Provides API methods for accessing LDAP authentication information."""

    def __init__(self, registry, manager, realm=None):
        self.registry = registry
        self.manager = manager
        self.realm = realm
        self.login_query_identifier = _add_realm('ldap_login_query', realm)
        self.group_query_identifier = _add_realm('ldap_groups_query', realm)

    def authenticate(self, login, password):
        """Validate the given login name and password.

        Given a login name and a password, return a tuple of ``(dn,
        attrdict)`` if the matching user if the user exists and his password
        is correct.  Otherwise return ``None``.

        In a ``(dn, attrdict)`` return value, ``dn`` will be the
        distinguished name of the authenticated user.  Attrdict will be a
        dictionary mapping LDAP user attributes to sequences of values.

        A zero length password will always be considered invalid since it
        results in a request for "unauthenticated authentication" which should
        not be used for LDAP based authentication. See `section 5.1.2 of
        RFC-4513 <https://tools.ietf.org/html/rfc4513#section-5.1.2>`_ for a
        description of this behavior.

        If :meth:`pyramid.config.Configurator.ldap_set_login_query` was not
        called, using this function will raise an
        :exc:`pyramid.exceptions.ConfiguratorError`.
        """

        if password == '':
            return None

        search = getattr(self.registry, self.login_query_identifier, None)
        if search is None:
            raise ConfigurationError(
                'ldap_set_login_query was not called during setup')

        result = search.execute(
            self.manager, login=escape_for_search(login),
            password=escape_for_search(password))

        if not result or len(result) > 1:
            return None
        result = result[0]
        login_dn = result[0]

        try:
            self.manager.connection(login_dn, password).unbind()
        except LDAPException:
            logger.debug(
                'Exception in authenticate with login %r', login,
                exc_info=True)
            return None

        return result

    def user_groups(self, userdn):
        """Get the groups the user belongs to.

        Given a user DN, return a sequence of LDAP attribute dictionaries
        matching the groups of which the DN is a member.  If the DN does not
        exist, return ``None``.

        In a return value ``[(dn, attrdict), ...]``, ``dn`` will be the
        distinguished name of the group.  Attrdict will be a dictionary
        mapping LDAP group attributes to sequences of values.

        If :meth:`pyramid.config.Configurator.ldap_set_groups_query` was not
        called, using this function will raise an
        :exc:`pyramid.exceptions.ConfiguratorError`

        """
        search = getattr(self.registry, self.group_query_identifier, None)
        if search is None:
            raise ConfigurationError(
                'set_ldap_groups_query was not called during setup')
        try:
            result = search.execute(
                self.manager, userdn=escape_for_search(userdn))
        except LDAPException:
            logger.debug(
                'Exception in user_groups with userdn %r', userdn,
                exc_info=True)
            return None

        return result


def ldap_set_login_query(
        config, base_dn, filter_tmpl,
        scope=ldap3.LEVEL, attributes=None,
        cache_period=0, realm=None):
    """Configurator method to set the LDAP login search.

    ``base_dn`` is the DN at which to begin the search.
    ``filter_tmpl`` is a string which can be used as an LDAP filter:
    it should contain the replacement value ``%(login)s``.
    ``scope`` is any valid LDAP scope value
    (e.g. ``ldap3.LEVEL`` or ``ldap3.SUBTREE``).
    ``attributes`` is an optional list of attributes that shall be returned
    (can also be set to None or ``ldap3.ALL_ATTRIBUTES``).
    ``cache_period`` is the number of seconds to cache login search results;
    if it is 0 (the default), login search results will not be cached.
    ``realm`` is an optional realm for this connection.
    This allows multiple ldap servers to be used.

    Example::

        config.set_ldap_login_query(
            base_dn='CN=Users,DC=example,DC=com',
            filter_tmpl='(sAMAccountName=%(login)s)',
            scope=ldap3.LEVEL)

    The registered search must return one and only one value to be considered
    a valid login.
    """
    query_identifier = _add_realm('ldap_login_query', realm)
    query = _LDAPQuery(base_dn, filter_tmpl, scope, attributes, cache_period)

    def register():
        setattr(config.registry, query_identifier, query)

    introspectable_name = '{} login query'.format(
        _add_realm('pyramid_ldap3', realm))
    introspectable = config.introspectable(
        introspectable_name,
        None,
        str(query),
        introspectable_name)

    action_name = 'setup_{}'.format(query_identifier)
    config.action(action_name, register,
                  introspectables=(introspectable,))


def ldap_set_groups_query(
        config, base_dn, filter_tmpl,
        scope=ldap3.SUBTREE, attributes=None,
        cache_period=0, realm=None):
    """ Configurator method to set the LDAP groups search.

    ``base_dn`` is the DN at which to begin the search.
    ``filter_tmpl`` is a string which can be used as an LDAP filter:
    it should contain the replacement value ``%(userdn)s``.
    ``scope`` is any valid LDAP scope value
    (e.g. ``ldap3.LEVEL`` or ``ldap3.SUBTREE``).
    ``attributes`` is an optional list of attributes that shall be returned
    (can also be set to None or ``ldap3.ALL_ATTRIBUTES``).
    ``cache_period`` is the number of seconds to cache groups search results;
    if it is 0 (the default), groups search results will not be cached.
    ``realm`` is an optional realm for this connection.
    This allows multiple ldap servers to be used.

    Example::

        config.set_ldap_groups_query(
            base_dn='CN=Users,DC=example,DC=com',
            filter_tmpl='(&(objectCategory=group)(member=%(userdn)s))'
            scope=ldap3.SUBTREE)

    """
    query_identifier = _add_realm('ldap_groups_query', realm)
    query = _LDAPQuery(base_dn, filter_tmpl, scope, attributes, cache_period)

    def register():
        setattr(config.registry, query_identifier, query)

    introspectable_name = '{} groups query'.format(
        _add_realm('pyramid_ldap3', realm))
    introspectable = config.introspectable(
        introspectable_name,
        None,
        str(query),
        introspectable_name)

    action_name = 'setup_{}'.format(query_identifier)
    config.action(action_name, register,
                  introspectables=(introspectable,))


def ldap_setup(
        config, uri,
        bind=None, passwd=None, use_tls=False,
        use_pool=True, pool_size=10, pool_lifetime=3600,
        get_info=None, realm=None):
    """Configurator method to set up an LDAP connection pool.

    - **uri**: ldap server uri(s) **[mandatory]**
    - **bind**: default bind that will be used to bind a connector.
      **default: None**
    - **passwd**: default password that will be used to bind a connector.
      **default: None**
    - **use_tls**: activate TLS when connecting. **default: False**
    - **use_pool**: activates the connection pool. If False, will recreate a
      connector each time. **default: True**
    - **pool_size**: connection pool size. **default: 10**
    - **pool_lifetime**: number of seconds before recreating a new connection
      when using a connection pool.  **default: 3600**
    - **get_info**: specifies if schema or server specific info shall be read
      for proper formatting of attributes.  **default: None**
    - **realm**: specify a realm for this connection.
      This allows multiple ldap servers to be used.  **default: None**
    """
    connection_identifier = _add_realm('ldap_connector', realm)

    manager = ConnectionManager(
        uri, bind, passwd, use_tls,
        use_pool, pool_size if use_pool else None,
        pool_lifetime if use_pool else None, get_info, realm=realm)

    def get_connector(request):
        return Connector(request.registry, manager, realm)

    config.add_request_method(
        get_connector, connection_identifier, property=True, reify=True)

    introspectable_name = '{} setup'.format(_add_realm('pyramid_ldap3', realm))
    introspectable = config.introspectable(
        introspectable_name,
        None,
        str(manager),
        introspectable_name)

    action_name = 'setup_{}'.format(connection_identifier)
    config.action(action_name, None, introspectables=(introspectable,))


def get_ldap_connector_name(realm=None):
    """Return the name of the connector attached to the request.

     You can also specify the ``realm`` for the connector.
     """
    return _add_realm('ldap_connector', realm)


def get_ldap_connector(request, realm=None):
    """Return the LDAP connector attached to the request.

    If :meth:`pyramid.config.Configurator.ldap_setup` was not called, using
    this function will raise an :exc:`pyramid.exceptions.ConfigurationError`.
    """
    conn_name = get_ldap_connector_name(realm)
    connector = getattr(request, conn_name, None)
    if connector is None:
        if LDAPException is Exception:  # pragma: no cover
            raise ImportError(
                'You must install ldap3 to use an LDAP connector.')
        raise ConfigurationError(
            'You must call Configurator.ldap_setup during setup '
            'to use an LDAP connector.')
    return connector


def get_groups(userdn, request):
    """Raw groupfinder function returning the complete group query result."""
    connector = get_ldap_connector(request)
    return connector.user_groups(userdn)


def groupfinder(userdn, request):
    """Groupfinder function for Pyramid.

    A groupfinder implementation useful in conjunction with out-of-the-box
    Pyramid authentication policies.  It returns the DN of each group
    belonging to the user specified by ``userdn`` to as a principal
    in the list of results; if the user does not exist, it returns None.
    """
    groups = get_groups(userdn, request)
    if groups:
        groups = [r[0] for r in groups]
    return groups


def includeme(config):
    """Set up Configurator methods for pyramid_ldap3."""
    config.add_directive('ldap_setup', ldap_setup)
    config.add_directive('ldap_set_login_query', ldap_set_login_query)
    config.add_directive('ldap_set_groups_query', ldap_set_groups_query)



from os import path
from setuptools import setup, find_packages

here = path.abspath(path.dirname(__file__))
README = open(path.join(here, 'README.rst')).read()
CHANGES = open(path.join(here, 'CHANGES.txt')).read()

requires = ['pyramid>=1.4', 'ldap3>=2.0']

sampleapp_extras = [
    'waitress', 'pyramid_chameleon', 'pyramid_debugtoolbar']
testing_extras = ['coverage']
docs_extras = [
    'Sphinx>=3,<4',
    'repoze.sphinx.autointerface>=0.8,<1',
    'pylons-sphinx-themes>=1.0.8']

setup(
    name='pyramid_ldap3',
    version='0.5',
    description='pyramid_ldap3',
    long_description=README + '\n\n' + CHANGES,
    long_description_content_type='text/x-rst',
    classifiers=[
        "Framework :: Pyramid",
        "Programming Language :: Python",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        'Topic :: Software Development :: Libraries :: Python Modules',
        "Topic :: System :: Systems Administration"
            " :: Authentication/Directory :: LDAP",
        "Intended Audience :: Developers",
        "Development Status :: 4 - Beta",
        "License :: Repoze Public License"],
    author='Chris McDonough, Christoph Zwerschke',
    author_email='pylons-discuss@googlegroups.com',
    url='https://trypyramid.com/extending-pyramid.html',
    license="BSD-derived (Repoze)",
    keywords=['web', 'pyramid', 'pylons', 'ldap', 'auth', 'authentication'],
    packages=find_packages(exclude=['docs', 'tests', 'sampleapp']),
    zip_safe=False,
    install_requires=requires,
    tests_require=requires,
    extras_require=dict(
        sampleapp=sampleapp_extras,
        docs=docs_extras,
        testing=testing_extras),
    test_suite='tests',
    entry_points={'paste.app_factory': ['sampleapp = sampleapp:main']})
